/*
 * Copyright (c) 2002, 2003
 *	The Regents of the University of California.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that: (1) source code distributions
 * retain the above copyright notice and this paragraph in its entirety, (2)
 * distributions including binary code include the above copyright notice and
 * this paragraph in its entirety in the documentation or other materials
 * provided with the distribution, and (3) all advertising materials mentioning
 * features or use of this software display the following acknowledgement:
 * ``This product includes software developed by the University of California,
 * Lawrence Berkeley Laboratory and its contributors.'' Neither the name of
 * the University nor the names of its contributors may be used to endorse
 * or promote products derived from this software without specific prior
 * written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */

#ifndef lint
static const char rcsid[] =
    "@(#) $Id: child.c,v 1.19 2003/04/10 01:54:05 leres Exp $ (LBL)";
#endif

#include <sys/types.h>
#include <sys/time.h>

#include <errno.h>
#ifdef HAVE_MALLOC_H
#include <malloc.h>
#endif
#include <signal.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sysexits.h>
#include <syslog.h>
#include <unistd.h>

#include "acld.h"
#include "cf.h"
#include "child.h"
#include "util.h"

/* Forwards */
int childresponse(char **, struct state *);
int failed(const char *, const char *);

/* Process input from the client */
void
childinput(register struct state *sp, register struct client *cl)
{
	register int an, flags, i;
	register char *cp;
	register struct acllist *ap, *ap2;
	register struct req *rp;
	char **av;
	static char *p1, *p2;
	register struct iobuf *ip;

	ip = &sp->childiobuf;
	switch (sp->state) {

	case ASTATE_READRESPONSE:
		if (childresponse(&p1, sp)) {
			if (sp->state == ASTATE_READRESPONSE) {
				sp->state = ASTATE_LOGGEDIN;
				sp->cmd = NULL;
			} else {
				sp->state = ASTATE_CONNECTED;
				sp->cmd = NULL;
			}
			if (p1 != NULL) {
				trimws(p1);
				log(LOG_DEBUG, "childinput response: \"%s\"",
				    pretty(p1, strlen(p1)));
				free(p1);
				p1 = NULL;
			}
		}
		return;

	case ASTATE_READERROR:
		if (childresponse(&p2, sp)) {
			if (sp->state == ASTATE_READRESPONSE) {
				sp->state = ASTATE_LOGGEDIN;
				sp->cmd = NULL;
			} else {
				sp->state = ASTATE_CONNECTED;
				sp->cmd = NULL;
				/* Wait a bit before trying again */
				timerset(&sp->t_login, sp->cf->c_login_secs);
			}
			if (p2 != NULL) {
				trimws(p2);
				log(LOG_DEBUG, "childinput error: \"%s\"",
				    pretty(p2, strlen(p2)));
				free(p2);
				p2 = NULL;
			}
		}
		return;

	case ASTATE_READACL:
		while ((cp = iogetline(ip)) != NULL) {
			/* Find the last acl list that hasn't been updated */
			ap2 = sp->cf->c_acllist;
			ap = NULL;
			for (i = 0; i < sp->cf->c_acllistlen; ++i) {
				if (!ap2->sentlistacl)
					break;
				ap = ap2++;
			}
			if (ap == NULL) {
				log(LOG_ERR, "childinput: can't happen");
				exit(EX_SOFTWARE);
			}

			/* Remove trailing newline for possible error message */
			trimws(cp);
			if (aclstraddacl(sp->cf, ap, cp)) {
				sp->state = ASTATE_LOGGEDIN;
				sp->cmd = NULL;
				break;
			}
		}
		return;

	default:
		if (cl != NULL && cl->req != NULL &&
		    cl->req->state == RSTATE_READRESPONSE) {
			rp = cl->req;
			while ((cp = iogetline(ip)) != NULL) {
				/* Insure proper trailing "\r\n" */
				trimws(cp);
				if (strcmp(cp, ".") == 0) {
					clientsendeom(sp, cl);
					rp->state = RSTATE_DONE;
					getts(&rp->cts);
					return;
				}
				log(LOG_DEBUG, "childinput response: \"%s\"",
				    pretty(cp, strlen(cp)));
				clientwrite(sp, cl, "%s", cp);
			}
			return;
		}
		break;
	}

	/* Break into arguments */
	cp = iogetline(ip);
	av = NULL;
	an = makeargv(cp, &av);

	if (an <= 0) {
		freeargv(av);
		return;
	}

	if (an == 1 && strcmp(av[0], "expect>") == 0) {
		if (sp->state == ASTATE_NOTCONNECTED) {
			sp->state = ASTATE_CONNECTED;
			sp->cmd = NULL;
		}
		if (debug)
			fprintf(stderr, "%s childinput: (ready)\n", tsstr());
		freeargv(av);
		return;
	}

	if (debug) {
		fprintf(stderr, "%s childinput:", tsstr());
		for (i = 0; i < an; ++i)
			fprintf(stderr, " %s", av[i]);
		putc('\n', stderr);
	}

	switch (sp->state) {

	case ASTATE_SENTAYT:
	case ASTATE_SENTLISTACL:
	case ASTATE_SENTLOGIN:
	case ASTATE_SENTSYNC:
		flags = 0;
		if (failed(sp->cmd, av[0]))
			flags |= CS_FAILED;
		else if (strcmp(sp->cmd, av[0]) != 0) {
			log(LOG_ERR,
			    "childinput: unexpected response %s waiting for %s",
			    av[0], sp->cmd);
			freeargv(av);
			return;
		}
		if (an > 1 && strcmp(av[1], "-") == 0)
			flags |= CS_CONTINUE;

		switch (sp->state) {

		case ASTATE_SENTLOGIN:
			if ((flags & CS_FAILED) == 0) {
				if ((flags & CS_CONTINUE) == 0)
					sp->state = ASTATE_LOGGEDIN;
				else
					sp->state = ASTATE_READRESPONSE;
			} else {
				if ((flags & CS_CONTINUE) == 0) {
					sp->state = ASTATE_CONNECTED;
					/* Wait a bit before trying again */
					timerset(&sp->t_login,
					    sp->cf->c_login_secs);
				} else
					sp->state = ASTATE_READERROR;
			}
			break;

		case ASTATE_SENTAYT:
			if ((flags & CS_FAILED) == 0) {
				if ((flags & CS_CONTINUE) == 0)
					sp->state = ASTATE_LOGGEDIN;
				else
					sp->state = ASTATE_READRESPONSE;
			} else {
				if ((flags & CS_CONTINUE) == 0)
					sp->state = ASTATE_NOTCONNECTED;
				else
					sp->state = ASTATE_READERROR;
			}
			break;

		case ASTATE_SENTSYNC:
			if ((flags & CS_FAILED) == 0) {
				if ((flags & CS_CONTINUE) == 0)
					sp->state = ASTATE_LOGGEDIN;
				else
					sp->state = ASTATE_READRESPONSE;
			} else {
				if ((flags & CS_CONTINUE) == 0)
					sp->state = ASTATE_LOGGEDIN;
				else
					sp->state = ASTATE_READERROR;
			}
			break;

		case ASTATE_SENTLISTACL:
			if ((flags & CS_FAILED) == 0) {
				if ((flags & CS_CONTINUE) == 0)
					sp->state = ASTATE_LOGGEDIN;
				else
					sp->state = ASTATE_READACL;
			} else {
				if ((flags & CS_CONTINUE) == 0) {
					sp->state = ASTATE_LOGGEDIN;
					childkill(sp);
				} else
					sp->state = ASTATE_READERROR;
			}
			break;

		default:
			break;
		}

		freeargv(av);
		return;

	default:
		break;
	}

	if (cl == NULL) {
		log(LOG_ERR, "childinput: no client");
		freeargv(av);
		return;
	}

	rp = cl->req;
	if (rp == NULL) {
		log(LOG_ERR, "childinput: no client request to process");
		freeargv(av);
		return;
	}

	if (rp->state != RSTATE_CHILD) {
		log(LOG_ERR, "childinput: not waiting for child to talk!");
		freeargv(av);
		return;
	}

	flags = 0;
	if (failed(rp->cmd, av[0]))
		flags = CS_FAILED;
	else if (strcmp(rp->cmd, av[0]) != 0) {
		log(LOG_ERR,
		    "childinput: unexpected response %s waiting for %s",
		    av[0], rp->cmd);
		freeargv(av);
		return;
	}

	if (an > 1 && strcmp(av[1], "-") == 0) {
		flags |= CS_CONTINUE;
		rp->state = RSTATE_READRESPONSE;
	} else {
		rp->state = RSTATE_DONE;
		getts(&rp->cts);
	}

	clientsend(sp, cl, rp->cookie, cl->req->cmd, flags);

	if ((flags & CS_FAILED) == 0) {
		switch (rp->type) {

		case REQ_DROP:
			ap = aclfindlistbyaddr(sp->cf, &rp->acl.addr1);
			if (ap == NULL) {
				log(LOG_ERR, "childinput: NULL %s", rp->cmd);
				exit(EX_SOFTWARE);
			}
			acladdacl(ap, &rp->acl);
			break;

		case REQ_RESTORE:
			ap = aclfindlistbyaddr(sp->cf, &rp->acl.addr1);
			if (ap == NULL) {
				log(LOG_ERR, "childinput: NULL %s", rp->cmd);
				exit(EX_SOFTWARE);
			}
			if (!acldeleteacl(ap, &rp->acl))
				log(LOG_ERR, "childinput: acldeleteacl failed");
			break;

		default:
			break;
		}
	}

	freeargv(av);
}

void
childayt(register struct state *sp)
{

	if (sp->wfd >= 0 && sp->state == ASTATE_LOGGEDIN) {
		sp->state = ASTATE_SENTAYT;
		sp->cmd = "ayt";
		childsend(sp, sp->cmd);
	}
}

void
childdroprestore(register struct state *sp, register struct client *cl,
    register struct req *rp)
{
	register struct acllist *ap;
	register struct acl *al;

	if (rp->acl.type != ATYPE_BLOCKHOST) {
		clientsenderr(sp, cl, rp->cookie, rp->cmd,
		    "don't know how to handle acl request type %d",
		    rp->acl.type);
		rp->state = RSTATE_DONE;
		getts(&rp->cts);
		return;
	}

	ap = aclfindlistbyaddr(sp->cf, &rp->acl.addr1);
	if (ap == NULL) {
		clientsenderr(sp, cl, rp->cookie, rp->cmd,
		    "can't find acllist for %s (is the default acl missing?)",
		    addr2str(rp->acl.addr1.addr));
		rp->state = RSTATE_DONE;
		getts(&rp->cts);
		return;
	}
	al = aclfindacl(ap, &rp->acl);

	switch (rp->type) {

	case REQ_DROP:
		if (al != NULL) {
			clientsend(sp, cl, rp->cookie, rp->cmd,
			    CS_CONTINUE);
			clientwrite(sp, cl,
			    "Note: %s is already blocked in acl %s",
			    addr2str(rp->acl.addr1.addr), ap->name);
			clientsendeom(sp, cl);
			rp->state = RSTATE_DONE;
			getts(&rp->cts);
			return;
		}
		if (sp->cf->c_highseq > 0 && ap->lastseq >= sp->cf->c_highseq) {
			clientsenderr(sp, cl, rp->cookie, rp->cmd,
			    "Last sequence number in ACL %s"
			    " already in use (%d)",
			    ap->name, ap->lastseq);
			rp->state = RSTATE_DONE;
			getts(&rp->cts);
			return;
		}
		rp->acl.seq = ++ap->lastseq;
		break;

	case REQ_RESTORE:
		if (al == NULL) {
			clientsenderr(sp, cl, rp->cookie, rp->cmd,
			    "%s not found in acl %s",
			    addr2str(rp->acl.addr1.addr), ap->name);
			getts(&rp->cts);
			rp->state = RSTATE_DONE;
			return;
		}
		rp->acl.seq = al->seq;
		break;

	default:
		log(LOG_ERR, "childdroprestore: bad type %d", rp->type);
		exit(EX_SOFTWARE);
	}

#ifdef notdef
	if (rp->acl.addr1.mask != 0xffffffff) {
		/* XXX Can't do networks yet */
		clientsenderr(sp, cl, rp->cookie, rp->cmd,
		    "childdroprestore: networks not implemented yet");
		getts(&rp->cts);
		rp->state = RSTATE_DONE;
		return;
	}
#endif

	childsend(sp, "%s %s %s %d",
	    rp->cmd, addr2str(rp->acl.addr1.addr), ap->name, rp->acl.seq);
	rp->state = RSTATE_CHILD;

	/* Do a sync about a little later */
	if (sp->cf->c_sync_secs > 0 && timercheck(&sp->t_sync) < 0)
		timerset(&sp->t_sync, sp->cf->c_sync_secs);
}

void
childkill(register struct state *sp)
{

	if (sp->wfd >= 0 && sp->state == ASTATE_LOGGEDIN)
		childsend(sp, "logout");
	sp->state = ASTATE_NOTCONNECTED;
	sp->cmd = NULL;

	errno = 0;
	if (sp->pid > 0 && kill(sp->pid, SIGTERM) < 0 && errno != ESRCH)
		log(LOG_ERR, "childkill: kill %d: %s",
		    (int)sp->pid, strerror(errno));

	sp->pid = -1;
	if (sp->rfd >= 0)
		close(sp->rfd);
	sp->rfd = -1;

	if (sp->wfd >= 0)
		close(sp->wfd);
	sp->wfd = -1;

	sp->nfds = 0;

	iofree(&sp->childiobuf);

	timerset(&sp->t_ayt, 0);
	timerset(&sp->t_login, 0);
	timerset(&sp->t_sync, 0);
}

void
childlistacl(register struct state *sp)
{
	register struct acllist *ap;
	register struct cf *cf;
	register int i;

	cf = sp->cf;
	for (i = 0, ap = cf->c_acllist; i < cf->c_acllistlen; ++i, ++ap)
		if (!ap->sentlistacl) {
			sp->state = ASTATE_SENTLISTACL;
			sp->cmd = "listacl";
			childsend(sp, "%s %s {%s}",
			    sp->cmd, ap->name, ap->intlist->i_name);
			++ap->sentlistacl;
			return;
		}

	/* If we got here, we've listed all the ACLs; sort 'em */
	++cf->c_listedallacls;
	aclsortacls(cf);
}

void
childlogin(register struct state *sp)
{
	register struct cf *cf;
	int odebug;

	/* Free acl lists so we'll re-aquire them from the router */
	cf = sp->cf;
	acllistsfree(cf);

	sp->state = ASTATE_SENTLOGIN;
	sp->cmd = "login";
	odebug = debug;
	if (odebug) {
		/* Avoid exposing router passwords in debugging output */
		debug = 0;
		fprintf(stderr, "%s childsend: \"%s %s %s %s %s %s %s %s\"\n",
		    tsstr(),
		    sp->cmd,
		    cf->c_router,
		    cf->c_cuser,
		    "?",
		    "?",
		    cf->c_euser,
		    "?",
		    "?");
	}
	childsend(sp, "%s %s %s %s %s %s %s %s",
	    sp->cmd,
	    cf->c_router,
	    cf->c_cuser,
	    cf->c_cpass1,
	    cf->c_cpass2,
	    cf->c_euser,
	    cf->c_epass1,
	    cf->c_epass2);
	debug = odebug;
	timerreset(&sp->t_ayt);
	timerreset(&sp->t_sync);
}

int
childresponse(register char **pp, register struct state *sp)
{
	register char *cp;
	register struct iobuf *ip;

	/* Look for end of message (and strip leading dots) */
	ip = &sp->childiobuf;
	while ((cp = iogetline(ip)) != NULL) {
		if (*cp == '.') {
			++cp;
			if (*cp == '\0' || *cp == '\n')
				return (1);
		}
		strappend(pp, cp);
	}
	return(0);
}

void
childsend(register struct state *sp, register const char *fmt, ...)
{
	register char *cp;
	register size_t n;
	register ssize_t cc;
	char buf[1024];
	va_list ap;

	va_start(ap, fmt);
	(void)vsnprintf(buf, sizeof(buf) - 1, fmt, ap);
	va_end(ap);
	n = strlen(buf);
	cp = buf + n - 1;

	if (debug)
		fprintf(stderr, "%s childsend: \"%s\"\n", tsstr(), buf);

	if (cp >= buf && *cp != '\n') {
		*++cp = '\n';
		*++cp = '\0';
		++n;
	}
	cc = write(sp->wfd, buf, n);
	if (cc < 0) {
		/* XXX close off child here? */
		log(LOG_ERR, "childsend: write: %s", strerror(errno));
		return;
	}
	if (cc != n) {
		log(LOG_ERR, "childsend: short write (%d != %d)", cc, n);
		return;
	}
}

void
childsync(register struct state *sp)
{

	sp->state = ASTATE_SENTSYNC;
	sp->cmd = "sync";
	childsend(sp, sp->cmd);
}

int
failed(register const char *what, register const char *str)
{
	register size_t len;

	len = strlen(what);
	if (strncmp(what, str, len) == 0 && strcmp(str + len, "-failed") == 0)
		return (1);
	return (0);
}
